"""
This module handles the history of migrations. It detects bifurcations.
Also comes with functionality to validate the tree.
"""

import importlib.util
import os
import re
import importlib

from typing import Dict, List


class MigrationNode:
    """
    Represents a migration node in the migration history tree.
    """

    def __init__(
        self,
        title: str,
        version: str,
        last_version: str = None,
        upgrade: str = None,
        downgrade: str = None,
    ):
        """
        Create a new migration node.
        Args:
            title: The title of the migration.
            version: The version of the migration.
            last_version: The last version of the migration. None if it is the first migration.
            upgrade: The upgrade function of the migration.
            downgrade: The downgrade function of the migration.
        """
        self.title = title
        self.version = version
        self.last_version = last_version
        self.children: List[MigrationNode] = []
        self._upgrade = upgrade
        self._downgrade = downgrade

    def add_child(self, child_node: "MigrationNode"):
        """
        Add a child node to the current node.
        Args:
            child_node: The child node to add.
        """
        self.children.append(child_node)

    def upgrade(self, db):
        """
        Apply the upgrade function of the migration.
        Args:
            db: The database to upgrade.
        """
        if self._upgrade is not None:
            self._upgrade(db)

    def downgrade(self, db):
        """
        Apply the downgrade function of the migration.
        Args:
            db: The database to downgrade.
        """
        if self._downgrade is not None:
            self._downgrade(db)

    @classmethod
    def from_file(cls, file_path: str) -> "MigrationNode":
        """
        Parse a migration file and return the migration node.
        Args:
            file_path: The path to the migration file.
        Raises:
            FileNotFoundError: If the file is not found.
            ValueError: If the file format is invalid.
        Returns:
            The migration node parsed from the file.
        """
        pattern = re.compile(
            r"""
            title:\s*(?P<title>.+)\n
            version:\s*(?P<version>\d+)\n
            last_version:\s*(?P<last_version>\d+|None)
        """,
            re.VERBOSE,
        )

        # May raise FileNotFoundError
        with open(file_path, "r") as f:
            content = f.read()
            match = pattern.search(content)
            if match is None:
                raise ValueError(f"Invalid migration file format on {file_path}.")
            title = match.group("title")
            version = match.group("version")
            last_version = match.group("last_version")
            last_version = None if last_version == "None" else last_version

            spec = importlib.util.spec_from_file_location("migration_module", file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            upgrade = getattr(module, "upgrade", None)
            downgrade = getattr(module, "downgrade", None)

            return cls(title, version, last_version, upgrade, downgrade)

    def __repr__(self):
        return (
            f"MigrationNode(title={self.title}, "
            f"version={self.version}, last_version={self.last_version})"
        )

    def __str__(self):
        return f"{self.version} - {self.title}"


class MigrationHistory:

    def __init__(self, migrations_dir: str):
        """
        Class for managing the migration history.
        Args:
            migrations_dir: The directory where the migrations are stored.
        Attributes:
            migrations_dir: The directory where the migrations are stored.
            roots: The root nodes of the migration history.
            migrations: A dictionary of all migrations by version.
        Raises:
            FileNotFoundError: If a migration file is not found.
            ValueError: If a migration file format is invalid.
        """
        self.migrations_dir = migrations_dir
        self.roots: List[MigrationNode] = []
        self.migrations: Dict[str, MigrationNode] = {}
        self._load_migrations()

    def _load_migrations(self):
        """
        Private method to load all migrations from the migrations directory.
        Migrations without a last version or without a found last_version are considered as roots.

        Raises:
            FileNotFoundError: If a migration file is not found.
            ValueError: If a migration file format is invalid.
        """
        migration_files = [
            f for f in os.listdir(self.migrations_dir) if f.endswith(".py")
        ]

        # Load all migration files
        for migration_file in migration_files:
            file_path = os.path.join(self.migrations_dir, migration_file)
            # May raise FileNotFoundError or ValueError
            node = MigrationNode.from_file(file_path)
            self.migrations[node.version] = node
            if node:
                self.migrations[node.version] = node

        # Build the tree
        for node in self.migrations.values():
            if all([node.last_version, node.last_version in self.migrations]):
                self.migrations[node.last_version].add_child(node)
            else:
                self.roots.append(node)

    def is_empty(self) -> bool:
        """
        Check if the migration history is empty.
        Returns:
            True if the history is empty, False otherwise.
        """
        return not self.roots

    def validate(self, node: MigrationNode = None) -> bool:
        """
        Check the migration history for bifurcations.
        The history is valid if it is a non empty linear tree.
        Args:
            node: The node to start the check. If None, the check starts from the root.
        Returns:
            True if the history is valid, False otherwise.
        """
        if self.is_empty():
            return False

        if node is None:
            # More than one root migration
            if len(self.roots) > 1:
                return False
            node = self.roots[0]

        # More than one child for a migration
        if len(node.children) > 1:
            return False

        # Recursive check
        return self.validate(node.children[0]) if node.children else True

    def get_first_version(self) -> str:
        """
        Get the first version of the migration history.
        Assumes that the migration history is valid.
        Returns:
            The first version of the migration history.
        """
        if self.is_empty():
            return None

        return self.roots[0].version

    def get_first_node(self) -> MigrationNode:
        """
        Get the first node of the migration history.
        Assumes that the migration history is valid.
        Returns:
            The first node of the migration history.
        """
        if self.is_empty():
            return None

        return self.roots[0]

    def get_last_version(self) -> str:
        """
        Get the last version of the migration history.
        Assumes that the migration history is valid.
        Returns:
            The last version of the migration history.
        """
        if self.is_empty():
            return None

        current_node = self.roots[0]
        while current_node.children:
            current_node = current_node.children[0]
        return current_node.version

    def get_last_node(self) -> MigrationNode:
        """
        Get the last node of the migration history.
        Assumes that the migration history is valid.
        Returns:
            The last node of the migration history.
        """
        if self.is_empty():
            return None

        current_node = self.roots[0]
        while current_node.children:
            current_node = current_node.children[0]
        return current_node

    def _print_linear_tree(
        self,
        node: MigrationNode,
        current_version: str = None,
        current_found: bool = False,
        is_applied: bool = True,
    ):
        """
        Print a linear tree of the migration history.
        Args:
            node: The node to print.
            level: The level of the node in the tree.
        """
        is_current = current_version == node.version
        current_found = True if is_current else current_found

        # If this is the current version, it is applied
        # Also, every following migration is not applied
        connection = "└──" if not node.children else "├──"
        if is_current:
            state = "(CURRENT) "
            connection += ">"
        else:
            state = " (APPLIED) " if is_applied else " (PENDING) "

        print(f"{connection}{state}{node}")

        if node.children:
            # All migrations before the current version are applied
            if not current_found:
                self._print_linear_tree(
                    node.children[0], current_version, current_found, is_applied
                )
            else:
                self._print_linear_tree(
                    node.children[0], current_version, current_found, False
                )

    def print_history(self, current_version: str = None):
        """
        Prints the history of migrations from oldest to newest.
        Args:
            current_version: The current version of the database.
        Raises:
            ValueError: If the history is invalid.
        """
        if not self.validate():
            raise ValueError("Invalid migration history.")

        first_node = self.get_first_node()

        self._print_linear_tree(
            # Start from the first node
            first_node,
            # Current version to compare
            current_version,
            # If the first node is the current version, the current version is found
            # Therefore, every following migration is not applied
            first_node.version == current_version,
            # If the current version is set, the first migration is applied
            current_version is not None,
        )

    def get_migrations(
        self, start_version: str = None, to_version: str = None
    ) -> List[MigrationNode]:
        """
        Get a list of migrations.
        Assumes that the migration history is valid.
        Args:
            start_version: The version to start from. If None, the first version is used.
            to_version: The last version of the migration list. If None, the last version is used.
        Returns:
            A list of migrations.
        """
        start_version = self.roots[0].version if not start_version else start_version
        to_version = self.get_last_version() if not to_version else to_version

        migrations = []
        version = start_version

        while version != to_version:
            # If the current version is not the target version, add the migration
            current_node = self.migrations[version]
            migrations.append(current_node)
            # If the current version has a child, update the current version
            if current_node.children:
                child = current_node.children[0]
                version = child.version
            else:
                break

        if version == to_version:
            migrations.append(self.migrations[to_version])

        return migrations



"""Module for database operations."""

from pymongo import MongoClient
from pymongo.database import Database


def get_db(
    db_host: str,
    db_port: int,
    db_name: str,
    db_user: str = None,
    db_pass: str = None,
    verbose: bool = False,
    max_retries: int = 3,
) -> Database:
    """
    Get the database connection using pymongo.
    Args:
        db_host: The hostname of the MongoDB server.
        db_port: The port number of the MongoDB server.
        db_name: The name of the database.
        db_user: The username for the database if needed.
        db_pass: The password for the database if needed.
        verbose: Whether to print messages.
        max_retries: The number of times to retry connecting to the database.
    Raises:
        Exception: If the connection cannot be established.
    Returns:
        The database connection.
    """
    retries = 0
    timeout_ms = 5000

    if verbose:
        print(f"Connecting to MongoDB at {db_host}:{db_port}...")

    while retries < max_retries:
        try:
            client = MongoClient(
                host=db_host,
                port=db_port,
                username=db_user,
                password=db_pass,
                serverSelectionTimeoutMS=timeout_ms,
                socketTimeoutMS=timeout_ms,
                connectTimeoutMS=timeout_ms,
            )
            db = client[db_name]
            db.list_collection_names()
            if verbose:
                print("[+] Connected to database.")
            return db
        except Exception as err:
            print(f"[F] Error connecting to database: {err}")
            retries += 1

    raise Exception(
        "Could not connect to database. Please check your connection details."
    )


def create_version_collection(db: Database, collection_name: str) -> None:
    """
    Create the version collection in the database.
    Args:
        db: The database connection.
        collection_name: The name of the collection.
    """
    if collection_name not in db.list_collection_names():
        db.create_collection(collection_name)
        print(f"[+] Created version collection '{collection_name}'")
    else:
        print(f"[+] Version collection '{collection_name}' already exists")

    if db[collection_name].count_documents({}) == 0:
        db[collection_name].insert_one({"current_version": None})
        print("[+] Inserted current version document.")
    else:
        print("[+] Current version document already exists.")


def set_current_version(db: Database, collection_name: str, version: str) -> None:
    """
    Set the current version.
    Args:
        db: The database connection.
        collection_name: The name of the version collection.
        version: The current version.
    """
    db[collection_name].update_one({}, {"$set": {"current_version": version}})


def get_current_version(db: Database, collection_name: str) -> str:
    """
    Get the current version.
    Args:
        db: The database connection.
        collection_name: The name of the version collection.
    Returns:
        The current version.
    """
    version = db[collection_name].find_one()
    if version:
        return version.get("current_version")



"""
Contains the utility to manage the configuration file.

Usage:
```
from config import Config
config = Config() # Loads the configuration
print(config.db_host) # Access the database host
print(config.db_port) # Access the database port
...
```

Available variables in the Config class implementation
"""

import configparser
import os


class Config:
    """Class to handle the configuration file"""

    CONFIG_FILE = "mongo-migrator.config"

    def __init__(self):
        """
        Initialize the Config class.

        Description:
            Loads the configuration file and provides access to the configuration values.
            This file must be named after CONFIG_FILE attribute.
            It also must be placed in the same directory where mongo-migrator is going to be used.

        Exit cases:
            If the configuration file is not found, the program will exit.
            If any of the required sections are missing, the program will exit.
        """
        self.config = configparser.ConfigParser()

        try:
            if not os.path.exists(self.CONFIG_FILE):
                raise FileNotFoundError()

            self.config.read(self.CONFIG_FILE)
            # Database configuration
            self.db_host = self.config.get("database", "host")
            self.db_port = self.config.getint("database", "port")
            self.db_name = self.config.get("database", "name")
            self.db_user = None
            self.db_password = None
            # Migrations configuration
            self.migrations_dir = self.config.get("migrations", "directory")
            self.mm_collection = self.config.get("migrations", "collection")
        except FileNotFoundError:
            print("[F] Configuration file not found.")
            print(
                "[F] Please create a configuration file named 'mongo-migrator.config'."
            )
            print("[F] Exiting...")
            exit(1)
        except configparser.NoSectionError as err:
            print("[F] Configuration file is missing sections.")
            print(err)
            print("[F] Exiting...")
            exit(1)

        try:
            self.db_user = self.config.get("database", "user")
        except configparser.NoOptionError:
            self.db_user = None

        try:
            self.db_password = self.config.get("database", "password")
        except configparser.NoOptionError:
            self.db_password = None



"""
Command line interface for the mongo migrator
"""

import argparse
import os
import sys

from datetime import datetime

from mongo_migrator import __version__
from mongo_migrator.config import Config
from mongo_migrator.db_utils import (
    get_db,
    create_version_collection,
    get_current_version,
    set_current_version,
)

from mongo_migrator.migration_template import MigrationTemplate
from mongo_migrator.migration_history import MigrationHistory


def init(args):
    """
    Needs a config file named 'mongo-migrator.config' in the current directory.
    Will create a migrations directory and a migration collection in the database.
    """
    print("[*] Initializing migrations...")
    print("[*] Reading mongo-migrator.config...")

    # May exit if cant be loaded
    config = Config()

    print("[*] Generating migration collection in the database...")
    try:
        db = get_db(
            config.db_host,
            config.db_port,
            config.db_name,
            config.db_user,
            config.db_password,
        )
        create_version_collection(db, config.mm_collection)
    except Exception as err:
        print(f"[F] Error connecting to the database: {err}")
        return

    print("[*] Generating migration directory...")
    if not os.path.exists(config.migrations_dir):
        os.makedirs(config.migrations_dir)
        print(f"[+] Migration directory created at: {config.migrations_dir}")
    else:
        print(f"[+] Migration directory already exists at: {config.migrations_dir}")

    print("[+] Mongo-Migrator initialization complete.")


def create(args):
    """
    Creates a new migration file in the migration directory.
    """
    print("[*] Creating a new migration file...")

    if not args.title:
        print("[F] Missing argument: <title>")
        return

    raw_title = args.title
    title = raw_title.lower().replace(" ", "_")

    # May exit if cant be loaded
    config = Config()

    # Check if the migrations directory exists
    if not os.path.exists(config.migrations_dir):
        print("[!] Migration directory not found.")
        print("[!] Run 'mongo-migrator init' to initialize the migrations.")
        print("[!] Run 'mongo-migrator create <title>' to create a new migration.")
        return

    date = datetime.now()
    version = date.strftime("%Y%m%d%H%M%S%f")
    file_name = f"{version}_{title}.py"
    migration_path = os.path.join(config.migrations_dir, file_name)

    try:
        migration_history = MigrationHistory(config.migrations_dir)
    except Exception as err:
        print(f"[F] Error loading the migration history: {err}")
        return
    if not migration_history.is_empty() and not migration_history.validate():
        print("[F] Migration history is not valid.")
        print("[F] Please fix the migration files before creating a new migration.")
        return
    last_version = migration_history.get_last_version()

    MigrationTemplate.create_migration_file(
        migration_path, raw_title, version, last_version
    )

    print(f"[+] Migration file created at: {migration_path}")


def upgrade(args):
    """
    Upgrades the database to the latest version by default.
    """
    # May exit if cant be loaded
    config = Config()

    # Check if the migrations directory exists
    if not os.path.exists(config.migrations_dir):
        print("[!] Migration directory not found.")
        print("[!] Run 'mongo-migrator init' to initialize the migrations.")
        print("[!] Run 'mongo-migrator create <title>' to create a new migration.")
        return

    # Get current version
    try:
        db = get_db(
            config.db_host,
            config.db_port,
            config.db_name,
            config.db_user,
            config.db_password,
        )
        current_version = get_current_version(db, config.mm_collection)
        new_current_version = current_version
    except Exception as err:
        print(f"[F] Error connecting to the database: {err}")
        return

    try:
        migration_history = MigrationHistory(config.migrations_dir)
    except Exception as err:
        print(f"[F] Error loading the migration history: {err}")
        return
    # Validate the migration history. Must be a linear tree
    if migration_history.is_empty():
        print("[F] No migrations found.")
        print("[F] Run 'mongo-migrator create <title>' to create a new migration.")
        return
    if not migration_history.validate():
        print("[F] Migration history is not valid.")
        print("[F] Please fix the migration files before upgrading the database.")
        return

    # If requested, upgrade to the specified version, else upgrade to the latest
    to_version = args.version if args and args.version else None
    print(f"[+] Current version: {current_version}")
    print(
        f"[*] Upgrading the database to version: {to_version if to_version else 'latest'}"
    )

    # Retrieve the migrations to run
    # These migrations start with the first one found (if no current_version is set)
    # and end with the specified version (if requested) or the latest one found.
    migrations = migration_history.get_migrations(current_version, to_version)
    # Avoid upgrading the current version since it is already up to date
    to_upgrade = [mig for mig in migrations if mig.version != current_version]

    if args.version:
        # Check the requested migration is included in the retrieved migrations
        for mig in to_upgrade:
            if mig.version == args.version:
                break
        else:
            print(f"[F] Migration {args.version} not found in the peniding migrations.")
            return

    # If there are no migrations to run, exit
    if not to_upgrade:
        print("[+] No migrations to run.")
        return

    # Run the migrations
    print(f"[*] Running {len(to_upgrade)} migrations...")
    success = 0
    try:
        for migration in to_upgrade:
            print(f"[*] Running migration: {migration}")
            migration.upgrade(db)
            new_current_version = migration.version
            success += 1
    except Exception as err:
        print(f"[F] Error running migrations: {err}")
        return
    finally:
        print(f"[+] {success}/{len(to_upgrade)} migrations run successfully.")
        # If the current version has changed after trying to run the migrations
        # set the new current version in the database
        if new_current_version != current_version:
            set_current_version(db, config.mm_collection, new_current_version)
            print(f"[+] Current version set to: {new_current_version}")


def downgrade(args):
    """
    Downgrades the database to the previous version by default.
    """
    # May exit if cant be loaded
    config = Config()

    if not os.path.exists(config.migrations_dir):
        print("[!] Migration directory not found.")
        print("[!] Run 'mongo-migrator init' to initialize the migrations.")
        print("[!] Run 'mongo-migrator create <title>' to create a new migration.")
        return

    # Get current version
    try:
        db = get_db(
            config.db_host,
            config.db_port,
            config.db_name,
            config.db_user,
            config.db_password,
        )
        current_version = get_current_version(db, config.mm_collection)
        new_current_version = current_version
    except Exception as err:
        print(f"[F] Error connecting to the database: {err}")
        return

    try:
        migration_history = MigrationHistory(config.migrations_dir)
    except Exception as err:
        print(f"[F] Error loading the migration history: {err}")
        return
    # Validate the migration history. Must be a linear tree
    if migration_history.is_empty():
        print("[F] No migrations found.")
        print("[F] Run 'mongo-migrator create <title>' to create a new migration.")
        return
    if not migration_history.validate():
        print("[F] Migration history is not valid.")
        print("[F] Please fix the migration files before upgrading the database.")
        return
    if current_version is None:
        print("[F] No migrations have been run yet.")
        return

    # If requested, downgrade to the specified version, else downgrade to the previous
    to_version = args.version if args and args.version else None
    full_downgrade = args.all

    print(f"[+] Current version: {current_version}")
    if full_downgrade:
        print("[*] Full downgrade requested.")
    elif to_version:
        print(f"[*] Downgrading the database to version: {to_version}")
    else:
        print("[*] Downgrading the database to the previous version")

    # Retrieve the migrations to run
    # Because its a backward operation, all the migrations from the first one to the current one
    # are retrieved in reverse order.
    migrations = migration_history.get_migrations(None, current_version)
    migrations = migrations[::-1]
    to_downgrade = []
    # Run downgrade all versions
    if full_downgrade:
        to_downgrade = migrations
    # Downgrade to the specified version (not downgrading that version)
    elif to_version:
        # First check if the requested version is in the migration history
        for migration in migrations:
            if migration.version == to_version:
                break
        else:
            print(f"[F] Migration {to_version} not found in the parent migrations.")
            return
        # Get the migrations to run
        for migration in migrations:
            if migration.version != to_version:
                to_downgrade.append(migration)
            else:
                break
    # Run downgrade on the current version to the previous one
    else:
        to_downgrade = [migrations[0]] if len(migrations) > 1 else []

    # If there are no migrations to run, exit
    if not to_downgrade:
        print("[+] No migrations to run.")
        return

    # Run the migrations
    print(f"[*] Running {len(to_downgrade)} migrations...")
    success = 0
    try:
        for migration in to_downgrade:
            print(f"[*] Running migration: {migration}")
            migration.downgrade(db)
            new_current_version = migration.last_version
            success += 1
    except Exception as err:
        print(f"[F] Error running migrations: {err}")
        return
    finally:
        print(f"[+] {success}/{len(to_downgrade)} migrations run successfully.")
        # If the current version has changed after trying to run the migrations
        # set the new current version in the database
        if new_current_version != current_version:
            set_current_version(db, config.mm_collection, new_current_version)
            print(f"[+] Current version set to: {new_current_version}")


def history(args):
    """Shows the migration history."""
    # May exit if cant be loaded
    config = Config()

    # Check if the migrations directory exists
    if not os.path.exists(config.migrations_dir):
        print("[!] Migration directory not found.")
        print("[!] Run 'mongo-migrator init' to initialize the migrations.")
        return

    # Get the current version
    try:
        db = get_db(
            config.db_host,
            config.db_port,
            config.db_name,
            config.db_user,
            config.db_password,
        )
        current_version = get_current_version(db, config.mm_collection)
    except Exception as err:
        print(f"[F] Error connecting to the database: {err}")
        return

    # Load the migration history
    try:
        migration_history = MigrationHistory(config.migrations_dir)
        print("[+] Migration history:")
        migration_history.print_history(current_version)
    except Exception as err:
        print(f"[F] Error loading the migration history: {err}")
        return


def main():
    parser = argparse.ArgumentParser(
        description="Command line interface for the mongo migrator"
    )
    subparsers = parser.add_subparsers()

    # Version and Help options
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help="show the version of mongo-migrator.",
    )

    # Subcommand: init
    parser_init = subparsers.add_parser("init", help="set up the mongo-migrator.")
    parser_init.description = init.__doc__
    parser_init.set_defaults(func=init)

    # Subcommand: create
    parser_create = subparsers.add_parser("create", help="create a new migration file.")
    parser_create.description = create.__doc__
    parser_create.add_argument("title", help="title of the migration")
    parser_create.set_defaults(func=create)

    # Subcommand: upgrade
    parser_upgrade = subparsers.add_parser(
        "upgrade", help="upgrade the database by running the migrations."
    )
    parser_upgrade.description = upgrade.__doc__
    parser_upgrade.add_argument(
        "--all", action="store_true", default=True, help="upgrade to the latest version"
    )
    parser_upgrade.add_argument(
        "--version", help="upgrade to the specified version using the timestamp."
    )
    parser_upgrade.set_defaults(func=upgrade)

    # Subcommand: downgrade
    parser_downgrade = subparsers.add_parser(
        "downgrade", help="downgrade the database by running the migrations."
    )
    parser_downgrade.description = downgrade.__doc__
    parser_downgrade.add_argument(
        "--single",
        action="store_true",
        default=True,
        help="downgrade to the previous version",
    )
    parser_downgrade.add_argument(
        "--all", action="store_true", help="downgrade all versions"
    )
    parser_downgrade.add_argument(
        "--version", help="downgrade to the specified version using the timestamp."
    )
    parser_downgrade.set_defaults(func=downgrade)

    # Subcommand: history
    parser_history = subparsers.add_parser(
        "history", help="show the migration history."
    )
    parser_history.description = history.__doc__
    parser_history.set_defaults(func=history)

    # Parse arguments
    args = parser.parse_args()

    # Ensure that func is set before calling it
    if hasattr(args, "func"):
        args.func(args)
    # If no subcommand is provided, show help message
    else:
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
This class handles the creation of migration files.

The create_migration_file method creates a new migration file with a template.
This template includes the title, version, and current version of the migration.
"""


class MigrationTemplate:

    TEMPLATE = (
        '"""\n'
        "title: {title}\n"
        "version: {version}\n"
        "last_version: {last_version}\n"
        '"""\n'
        "from pymongo.database import Database\n"
        "\n"
        "def upgrade(db: Database):\n"
        "    # Implement this method in the generated migration file\n"
        "    pass\n"
        "\n"
        "def downgrade(db: Database):\n"
        "    # Implement this method in the generated migration file\n"
        "    pass\n"
        ""
    )

    @classmethod
    def create_migration_file(
        cls, file_path: str, title: str, version: int, last_version: int
    ):
        """
        Create a new migration file.
        Args:
            file_path: The path to the new migration file.
            title: The title of the migration.
            version: The version of the migration.
            last_version: The oldest version of the migrations.
        """
        with open(file_path, "w") as file:
            file.write(
                cls.TEMPLATE.format(
                    title=title, version=version, last_version=last_version
                )
            )
